'''Various utilities used here and there in the Foliant code.'''

from contextlib import contextmanager
from pkgutil import iter_modules
from importlib import import_module
from shutil import rmtree
from traceback import format_exc
from pathlib import Path
from logging import Logger
from typing import Dict, Tuple, Type, Set


def get_available_tags() -> Set[str]:
    '''Extract ``tags`` attribute values from installed
    ``foliant.preprocessors.*.Preprocessor`` classes.

    :returns: Set of tags
    '''

    preprocessors_module = import_module('foliant.preprocessors')

    result = set()

    for importer, modname, _ in iter_modules(preprocessors_module.__path__):
        if modname == 'base':
            continue

        result.update(importer.find_module(modname).load_module(modname).Preprocessor.tags)

    return result


def get_available_config_parsers() -> Dict[str, Type]:
    '''Get the names of the installed ``foliant.config`` submodules and the corresponding
    ``Parser`` classes.

    Used for construction of the Foliant config parser, which is a class that inherits
    from all ``foliant.config.*.Parser`` classes.

    :returns: Dictionary with submodule names as keys as classes as values
    '''

    config_module = import_module('foliant.config')

    result = {}

    for importer, modname, _ in iter_modules(config_module.__path__):
        if modname == 'base':
            continue

        result[modname] = importer.find_module(modname).load_module(modname).Parser

    return result


def get_available_clis() -> Dict[str, Type]:
    '''Get the names of the installed ``foliant.cli`` submodules and the corresponding
    ``Cli`` classes.

    Used for construction of the Foliant CLI, which is a class that inherits
    from all ``foliant.cli.*.Cli`` classes.

    :returns: Dictionary with submodule names as keys as classes as values
    '''

    cli_module = import_module('foliant.cli')

    result = {}

    for importer, modname, _ in iter_modules(cli_module.__path__):
        if modname == 'base':
            continue

        result[modname] = importer.find_module(modname).load_module(modname).Cli

    return result


def get_available_backends() -> Dict[str, Tuple[str]]:
    '''Get the names of the installed ``foliant.backends`` submodules and the corresponding
    ``Backend.targets`` tuples.

    Used in the interactive backend selection prompt to list the available backends
    and to check if the selected target can be made with the selected backend.

    :returns: Dictionary of submodule names as keys and target tuples as values

    '''

    backends_module = import_module('foliant.backends')

    result = {}

    for importer, modname, _ in iter_modules(backends_module.__path__):
        if modname == 'base':
            continue

        result[modname] = importer.find_module(modname).load_module(modname).Backend.targets

    return result


def output(text: str, quiet=False):
    '''Outputs the text to STDOUT in non-quiet mode

    :param text: Message to output
    '''

    if not quiet:
        print(text)


@contextmanager
def spinner(text: str, logger: Logger, quiet=False, debug=False):
    '''Decoration for long running processes.

    :param text: Message to output
    :param logger: Logger to capture the error if it occurs
    :param quiet: If ``True``, messages will be hidden
    :param debug: If ``True``, show full tracebacks
    '''

    # pylint: disable=broad-except

    try:
        logger.info(text)

        if not quiet:
            print(text, end='... ')

        yield

        if not quiet:
            print('Done')

    except Exception as exception:
        exception_traceback = format_exc()

        logger.error(exception_traceback)

        if not quiet:
            if debug:
                print(exception_traceback)

            else:
                print(str(exception))


@contextmanager
def tmp(tmp_path: Path, keep_tmp=False):
    '''Clean up tmp directory before and after running a code block.

    :param tmp_path: Path to the tmp directory
    :param keep_tmp: If ``True``, skip the cleanup
    '''

    rmtree(tmp_path, ignore_errors=True)

    yield

    if not keep_tmp:
        rmtree(tmp_path, ignore_errors=True)



from pathlib import Path

from yaml import load, add_constructor, Loader

from foliant.config.base import BaseParser


class Parser(BaseParser):
    def _resolve_include_tag(self, _, node) -> str:
        '''Replace value after ``!include`` with the content of the referenced file.'''

        parts = node.value.split('#')

        if len(parts) == 1:
            path = Path(parts[0]).expanduser()

            with open(self.project_path/path) as include_file:
                return load(include_file, Loader)

        elif len(parts) == 2:
            path, section = Path(parts[0]).expanduser(), parts[1]

            with open(self.project_path/path) as include_file:
                return load(include_file, Loader)[section]

        else:
            raise ValueError('Invalid include syntax')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        add_constructor('!include', self._resolve_include_tag)



from pathlib import Path
from logging import Logger

from yaml import load, Loader


class BaseParser(object):
    _defaults = {
        'src_dir': Path('./src'),
        'tmp_dir': Path('./__folianttmp__')
    }

    def __init__(
            self,
            project_path: Path,
            config_file_name: str,
            logger: Logger,
            quiet: bool = False
    ):
        self.project_path = project_path
        self.config_path = project_path / config_file_name
        self.logger = logger.getChild('cfg')
        self.quiet = quiet

    def parse(self) -> dict:
        '''Parse the config file into a Python dict. Missing values are populated
        with defaults, paths are converted to ``pathlib.Paths``.

        :param project_path: Project path
        :param config_file_name: Config file name (almost certainly ``foliant.yml``)

        :returns: Dictionary representing the YAML tree
        '''

        self.logger.info('Parsing started.')

        with open(self.config_path, encoding='utf8') as config_file:
            config = {**self._defaults, **load(config_file, Loader)}

            config['src_dir'] = Path(config['src_dir']).expanduser()
            config['tmp_dir'] = Path(config['tmp_dir']).expanduser()

            self.logger.info(f'Parsing completed.')

            self.logger.debug(f'Config: {config}')

            if not config.get('escape_code', False):
                self.logger.warning(
                    'Working in backward compatibility mode. ' +
                    'To get rid of this warning, enable the escape_code config option'
                )

            return config



from importlib import import_module
from shutil import copytree
from datetime import date
from logging import Logger

from foliant.utils import spinner


class BaseBackend(object):
    '''Base backend. All backends must inherit from this one.'''

    targets = ()
    required_preprocessors_before = ()
    required_preprocessors_after = ()

    def __init__(self, context: dict, logger: Logger, quiet=False, debug=False):
        self.project_path = context['project_path']
        self.config = context['config']
        self.context = context
        self.logger = logger
        self.quiet = quiet
        self.debug = debug

        self.working_dir = self.project_path / self.config['tmp_dir']

    def get_slug(self) -> str:
        '''Generate a slug from the project title and version and the current date.

        Spaces in title are replaced with underscores, then the version and the current date
        are appended.
        '''

        if 'slug' in self.config:
            return self.config['slug']

        components = []

        components.append(self.config['title'].replace(' ', '_'))

        version = self.config.get('version')
        if version:
            components.append(str(version))

        components.append(str(date.today()))

        return '-'.join(components)

    def apply_preprocessor(self, preprocessor: str or dict):
        '''Apply preprocessor.

        :param preprocessor: Preprocessor name or a dict of the preprocessor name and its options
        '''

        if isinstance(preprocessor, str):
            preprocessor_name, preprocessor_options = preprocessor, {}
        elif isinstance(preprocessor, dict):
            (preprocessor_name, preprocessor_options), = (*preprocessor.items(),)

        with spinner(
            f'Applying preprocessor {preprocessor_name}',
            self.logger,
            self.quiet,
            self.debug
        ):
            try:
                preprocessor_module = import_module(f'foliant.preprocessors.{preprocessor_name}')
                preprocessor_module.Preprocessor(
                    self.context,
                    self.logger,
                    self.quiet,
                    self.debug,
                    preprocessor_options
                ).apply()

            except ModuleNotFoundError:
                raise ModuleNotFoundError(f'Preprocessor {preprocessor_name} is not installed')

            except Exception as exception:
                raise type(exception)(
                    f'Failed to apply preprocessor {preprocessor_name}: {exception}'
                )

    def preprocess_and_make(self, target: str) -> str:
        '''Apply preprocessors required by the selected backend and defined in the config file,
        then run the ``make`` method.

        :param target: Output format: pdf, docx, html, etc.

        :returns: Result as returned by the ``make`` method
        '''

        src_path = self.project_path / self.config['src_dir']

        copytree(src_path, self.working_dir)

        common_preprocessors = (
            *self.required_preprocessors_before,
            *self.config.get('preprocessors', ()),
            *self.required_preprocessors_after
        )

        if self.config.get('escape_code', False):
            if isinstance(self.config['escape_code'], dict):
                escapecode_preprocessor = {
                    'escapecode': self.config['escape_code'].get('options', {})
                }

            else:
                escapecode_preprocessor = 'escapecode'

            preprocessors = (
                escapecode_preprocessor,
                *common_preprocessors,
                'unescapecode'
            )

        else:
            preprocessors = (
                *common_preprocessors,
                '_unescape'
            )

        for preprocessor in preprocessors:
            self.apply_preprocessor(preprocessor)

        return self.make(target)

    def make(self, target: str) -> str:
        '''Make the output from the source. Must be implemented by every backend.

        :param target: Output format: pdf, docx, html, etc.

        :returns: Typically, the path to the output file, but in general any string
        '''

        raise NotImplementedError



from shutil import copytree, rmtree

from foliant.backends.base import BaseBackend


class Backend(BaseBackend):
    '''Backend that just applies its preprocessors and returns a project
    that doesn't need any further preprocessing.
    '''

    targets = ('pre',)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._pre_config = self.config.get('backend_config', {}).get('pre', {})

        self._preprocessed_dir_name = f'{self._pre_config.get("slug", self.get_slug())}.pre'

        self.logger = self.logger.getChild('pre')

        self.logger.debug(f'Backend inited: {self.__dict__}')

    def make(self, target: str) -> str:
        rmtree(self._preprocessed_dir_name, ignore_errors=True)
        copytree(self.working_dir, self._preprocessed_dir_name)

        return self._preprocessed_dir_name



from pathlib import Path

from yaml import add_constructor

from foliant.config.base import BaseParser


class Parser(BaseParser):
    def _resolve_path_tag(self, _, node) -> str:
        '''Convert value after ``!path`` to an existing, absolute Posix path.

        Relative paths are relative to the project path.
        '''

        path = Path(node.value).expanduser()
        return (self.project_path / path).resolve(strict=True).as_posix()

    def _resolve_project_path_tag(self, _, node) -> str:
        '''Convert value after ``!project_path`` to Path object relative to the
        project path.

        Return absolute path to this file without checks for existance.
        '''

        return (self.project_path / node.value).resolve()

    @staticmethod
    def _resolve_rel_path_tag(_, node) -> str:
        '''Convert value after ``!rel_path`` to Path object.'''

        return Path(node.value)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        add_constructor('!path', self._resolve_path_tag)
        add_constructor('!project_path', self._resolve_project_path_tag)
        add_constructor('!rel_path', self._resolve_rel_path_tag)



from time import time
from logging import getLogger, FileHandler, Formatter

from cliar import set_help

from foliant import __version__ as foliant_version
from foliant.utils import get_available_clis


class Foliant(*get_available_clis().values()):
    '''Foliant is an all-in-one modular documentation authoring tool.'''

    @set_help({'version': 'show version and exit'})
    def _root(self, version=False):
        # pylint: disable=no-self-use

        if version:
            print(f'Foliant v.{foliant_version}')

        else:
            self._parser.print_help()


def entry_point():
    Foliant().parse()



from time import time
from logging import getLogger, FileHandler, Formatter

from cliar import Cliar


class BaseCli(Cliar):
    '''Base CLI. All CLI extensions must inherit from this one.'''

    def __init__(self):
        super().__init__()
        self.logger = getLogger('flt')
        handler = FileHandler(f'{int(time())}.log', delay=True)
        handler.setFormatter(Formatter('%(asctime)s | %(name)20s | %(levelname)8s | %(message)s'))
        self.logger.addHandler(handler)



import re
from logging import Logger
from typing import Dict
import yaml
OptionValue = int or float or bool or str


class BasePreprocessor(object):
    '''Base preprocessor. All preprocessors must inherit from this one.'''

    # pylint: disable=too-many-instance-attributes

    defaults = {}
    tags = ()

    @staticmethod
    def get_options(options_string: str) -> Dict[str, OptionValue]:
        '''Get a dictionary of typed options from a string with XML attributes.

        :param options_string: String of XML attributes

        :returns: Dictionary with options
        '''

        if not options_string:
            return {}

        option_pattern = re.compile(
            r'(?P<key>[A-Za-z_:][0-9A-Za-z_:\-\.]*)=(\'|")(?P<value>.+?)\2',
            flags=re.DOTALL
        )

        return {
            option.group('key'): yaml.load(option.group('value'), yaml.Loader)
            for option in option_pattern.finditer(options_string)
        }

    def __init__(self, context: dict, logger: Logger, quiet=False, debug=False, options={}):
        # pylint: disable=dangerous-default-value
        # pylint: disable=too-many-arguments

        self.project_path = context['project_path']
        self.config = context['config']
        self.context = context
        self.logger = logger
        self.quiet = quiet
        self.debug = debug
        self.options = {**self.defaults, **options}

        self.working_dir = self.project_path / self.config['tmp_dir']

        if self.tags:
            self.pattern = re.compile(
                rf'(?<!\<)\<(?P<tag>{"|".join(self.tags)})' +
                rf'(\s(?P<options>[^\<\>]*))?\>' +
                rf'(?P<body>.*?)\<\/(?P=tag)\>',
                flags=re.DOTALL
            )

    def apply(self):
        '''Run the preprocessor against the project directory. Must be implemented
        by every preprocessor.
        '''

        raise NotImplementedError



'''Main CLI repsonsible for the ``make`` command.'''

from pathlib import Path
from importlib import import_module
from logging import DEBUG, WARNING
from typing import List, Dict, Tuple

from cliar import set_arg_map, set_metavars, set_help, ignore
from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.validation import Validator, ValidationError

from foliant.config import Parser
from foliant.utils import spinner, get_available_backends, tmp
from foliant.cli.base import BaseCli


class ConfigError(Exception):
    pass


class BackendError(Exception):
    pass


class BackendValidator(Validator):
    '''Validator for the interactive backend selection prompt.'''

    def __init__(self, available_backends: List[str]):
        super().__init__()
        self.available_backends = available_backends

    def validate(self, document):
        '''Check if the selected backend in installed.'''

        backend = document.text

        if backend not in self.available_backends:
            raise ValidationError(
                message=f'Backend {backend} not found. '
                + f'Available backends are: {", ".join(self.available_backends)}.',
                cursor_position=0
            )


class Cli(BaseCli):
    @staticmethod
    def validate_backend(backend: str, target: str) -> bool:
        '''Check that the specified backend exists and can build the specified target.'''

        available_backends = get_available_backends()

        if backend not in available_backends:
            raise BackendError(
                f'Backend {backend} not found. '
                + f'Available backends are {", ".join(available_backends.keys())}.'
            )

        if target not in available_backends[backend]:
            raise BackendError(f'Backend {backend} cannot make {target}.')

        return True

    @staticmethod
    def get_matching_backend(target: str, available_backends: Dict[str, Tuple[str]]) -> str:
        '''Get a matching backend for the specified target. If multiple backends match
        the specified target, prompt user.'''

        matching_backends = [
            backend
            for backend, targets in available_backends.items()
            if target in targets
        ]

        if not matching_backends:
            raise BackendError(f'No backend available for {target}.')

        if len(matching_backends) == 1:
            return matching_backends[0]

        try:
            return prompt(
                f'Please pick a backend from {matching_backends}: ',
                completer=WordCompleter(matching_backends),
                validator=BackendValidator(matching_backends)
            )
        except KeyboardInterrupt:
            raise BackendError('No backend specified')

    @ignore
    def get_config(
            self,
            project_path: Path,
            config_file_name: str,
            quiet=False,
            debug=False
    ) -> dict:
        with spinner('Parsing config', self.logger, quiet, debug):
            try:
                config = Parser(project_path, config_file_name, self.logger, quiet).parse()

            except FileNotFoundError as exception:
                config = None
                raise FileNotFoundError(f'{exception} not found')

            except Exception as exception:
                config = None
                raise type(exception)(f'Invalid config: {exception}')

        if config is None:
            raise ConfigError('Config parsing failed.')

        return config

    @set_arg_map({'backend': 'with', 'project_path': 'path', 'config_file_name': 'config'})
    @set_metavars({'target': 'TARGET', 'backend': 'BACKEND'})
    @set_help(
        {
            'target': 'Target format: pdf, docx, html, etc.',
            'backend': 'Backend to make the target with: Pandoc, MkDocs, etc.',
            'project_path': 'Path to the Foliant project.',
            'config_file_name': 'Name of config file of the Foliant project.',
            'quiet': 'Hide all output accept for the result. Useful for piping.',
            'keep_tmp': 'Keep the tmp directory after the build.',
            'debug': 'Log all events during build. If not set, only warnings and errors are logged.'
        }
    )
    def make(
            self,
            target,
            backend='',
            project_path=Path('.'),
            config_file_name='foliant.yml',
            quiet=False,
            keep_tmp=False,
            debug=False
        ):
        '''Make TARGET with BACKEND.'''

        # pylint: disable=too-many-arguments

        self.logger.setLevel(DEBUG if debug else WARNING)

        self.logger.info('Build started.')

        available_backends = get_available_backends()

        try:
            if backend:
                self.validate_backend(backend, target)
            else:
                backend = self.get_matching_backend(target, available_backends)

            config = self.get_config(project_path, config_file_name, quiet, debug)

        except (BackendError, ConfigError) as exception:
            self.logger.critical(str(exception))
            exit(str(exception))

        context = {
            'project_path': project_path,
            'config': config,
            'target': target,
            'backend': backend
        }

        backend_module = import_module(f'foliant.backends.{backend}')
        self.logger.debug(f'Imported backend {backend_module}.')

        with tmp(project_path/config['tmp_dir'], keep_tmp):
            result = backend_module.Backend(
                context,
                self.logger,
                quiet,
                debug
            ).preprocess_and_make(target)

        if result:
            self.logger.info(f'Result: {result}')

            if not quiet:
                print('─' * 20)
                print(f'Result: {result}')
            else:
                print(result)

        else:
            self.logger.critical('No result returned by backend')
            exit('No result returned by backend')
            return None

        return result



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['foliant',
 'foliant.backends',
 'foliant.cli',
 'foliant.config',
 'foliant.preprocessors']

package_data = \
{'': ['*']}

install_requires = \
['cliar>=1.3.2,<2.0.0', 'prompt_toolkit>=2.0,<3.0', 'pyyaml>=5.1.1,<6.0.0']

entry_points = \
{'console_scripts': ['foliant = foliant.cli:entry_point']}

setup_kwargs = {
    'name': 'foliant',
    'version': '1.0.11',
    'description': 'Modular, Markdown-based documentation generator that makes pdf, docx, html, and more.',
    'long_description': '[![PyPI](https://img.shields.io/pypi/v/foliant.svg)](https://pypi.org/project/foliant/)\n[![Travis](https://img.shields.io/travis/foliant-docs/foliant.svg)](https://travis-ci.org/foliant-docs/foliant)\n[![codecov](https://codecov.io/gh/foliant-docs/foliant/branch/develop/graph/badge.svg)](https://codecov.io/gh/foliant-docs/foliant)\n\n# Foliant\n\nFoliant is a all-in-one documentation authoring tool. It lets you produce standalone documents in pdf and docx, as well as websites, from single Markdown source.\n\nFoliant is a *higher order* tool, which means that it uses other programs to do its job. For pdf and docx, it uses [Pandoc](http://pandoc.org/), for websites it uses [MkDocs](http://www.mkdocs.org/).\n\nFoliant preprocessors let you include parts of documents in other documents, show and hide content with flags, render diagrams from text, and more.\n\n## Installation\n\nFoliant is written in Python and requires Python 3.6.\n\nTo install Foliant, use pip:\n\n```shell\n$ pip install foliant\n```\n\nBackends, extensions, and preprocessors are installed with pip as well. For the Quickstart, you\'ll need the [MkDocs backend]() and [init CLI extension]():\n\n```shell\n$ pip install foliantcontrib.mkdocs foliantcontrib.init\n```\n\nTo build pdf and docx, install the [Pandoc backend]():\n\n```shell\n$ pip install foliantcontrib.pandoc\n```\n\n\n## Quickstart\n\n1. Create a new project:\n\n```shell\n$ foliant init\nEnter the project name: Hello Foliant\n✓ Generating Foliant project\n─────────────────────\nProject "Hello Foliant" created in /path/to/hello-foliant\n```\n\nThis command creates a basic Foliant project:\n\n```\nhello-foliant/\n├── foliant.yml\n└── src\n    └── index.md\n\n1 directory, 2 files\n```\n\n`foliant.yml` is the project config file, `src` is the directory that contains the project source files (initially, just one file `index.md`).\n\n2. Build a website from the newly created project:\n\n```shell\n$ foliant make site -p hello-foliant/\n✓ Parsing config\n✓ Applying preprocessor mkdocs\n✓ Making site with MkDocs\n─────────────────────\nResult: Hello_Foliant-0.1.0-2017-11-24.mkdocs\n```\n\n3. Run a local webserver in the site directory and see the site in your browser:\n\n```shell\n$ cd Hello_Foliant-0.1.0-2017-11-24.mkdocs\n$ python -m http.server\nServing HTTP on 0.0.0.0 port 8000 (http://0.0.0.0:8000/) ...\n```\n\n4. Build a standalone pdf document from the project:\n\n```shell\n$ foliant make pdf -p hello-foliant/\n✓ Parsing config\n✓ Applying preprocessor mkdocs\n✓ Making pdf with Pandoc\n─────────────────────\nResult: Hello_Foliant-0.1.0-2017-11-24.pdf\n```\n\n> **Important**\n>\n> To produce pdf, Pandoc first converts Markdown to tex and then coverts it to pdf. To convert pdf > from tex, you need to have a LaTeX distribution installed:\n>\n> -   [MacTeX](http://tug.org/mactex/) for macOS\n> -   [MikTeX](https://miktex.org/) for Windows\n> -   [TeXLive](https://tug.org/texlive/) for Linux\n',
    'author': 'Konstantin Molchanov',
    'author_email': 'moigagoo@live.com',
    'url': 'https://foliant-docs.github.io/docs/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import re

from foliant.preprocessors.base import BasePreprocessor


class Preprocessor(BasePreprocessor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.pattern = re.compile(
            rf'\<\<(?P<tag>[^\<\>\s]+)' +
            rf'(\s(?P<options>[^\<\>]*))?\>' +
            rf'(?P<body>.*?)\<\/(?P=tag)\>',
            flags=re.DOTALL
        )

        self.logger = self.logger.getChild('_unescape')

        self.logger.debug(f'Preprocessor inited: {self.__dict__}')

    def process_escaped_tags(self, content: str) -> str:
        '''Unescape escaped tags, i.e. remove leading ``<`` from each tag definition.

        :param content: Markdown content

        :returns: Markdown content without escaped tags
        '''

        def _sub(escaped_tag):
            tag_group = escaped_tag.group(0)
            result = tag_group[1:]

            self.logger.debug(
                f'Replacing {tag_group} with {result}'
            )

            return result

        return self.pattern.sub(_sub, content)

    def apply(self):
        for markdown_file_path in self.working_dir.rglob('*.md'):
            self.logger.debug(f'Processing the file: {markdown_file_path}')

            with open(markdown_file_path, encoding='utf8') as markdown_file:
                content = markdown_file.read()

            with open(markdown_file_path, 'w', encoding='utf8') as markdown_file:
                markdown_file.write(self.process_escaped_tags(content))
